#!/usr/bin/python
# -*-coding: utf-8-*-

import numpy as np
import scipy.optimize as scopt
import math
import matplotlib.pyplot as plt

# Programm zur Bereitstellung 
# der Funktion f(y,x,t) der
# DAE f(x',x,t) = 0.
def f(y,x,t):
    f1 = y[0] - x[1]
    f2 = y[1] - x[2]
    f3 = x[0] - math.sin(t)
    return np.array([f1,f2,f3])
# endfunction

# Programm 9.1  (bdf2.py)
# zur Loesung von gew. DGL 1. Ordnung f(x',x,t)=0, x(t0)=x0
# 2-schrittiges BDF-Verfahren (Ordnung 2)
# input:  Schrittweite h, Intervallgrenzen t0,tf,
#         Anfangswerte x0=x(t0), x1=x(t0+h)
# output: Loesung x(t) an den Punkten t
# benutztes Programm: Nutzerfunktion f (DAE Funktion)
# Aufruf: [x,t] = bdf2(t0,tf,x0,x1,h);

def bdf2(t0,tf,x0,x1,h):
    n = int((tf - t0)/h)
    m = len(x0)

    t = np.zeros(n+1); x = np.zeros((n+1,m))
    xk = np.zeros(m);
    xkm1 = np.zeros(m)
    t[0] = t0; t[1] = t0+h
    x[0] = x0; x[1] = x1
    xkm1 = x0 # starte mit x_{k-1}:=x0
    xk = x1   # und x_{k}:=x1
    tkp1 = t0+2.0*h # neuer Zeitpunkt t_{k+1}

    # Die Loesung xkp1=x_{k+1} eines BDF-Schrittes
    # ist die Nullstelle von bdf2step_f (nahe x_{k}).
    def bdf2step_f(x,xk,xkm1,tkp1,h):
        ykp1 = (1.5*x-2.0*xk+0.5*xkm1)/h
        return f(ykp1,x,tkp1)

    for k in range(1,n):
        args = (xk,xkm1,tkp1,h)

        # berechne Nullstelle von bdf2step_f
        # und nehme x_{k} als Startnaeherung:
#        xkp1 = scopt.fsolve(bdf2step_f,xk,args)

        # berechne Nullstelle direkt als Loesung eines lin. GLS mit einer oberen Dreiecksmatrix
        a = math.sin(tkp1)
        b = (1.5*a - 2*xk[0] + 0.5*xkm1[0])/h
        c = (1.5*b - 2*xk[1] + 0.5*xkm1[1])/h

        xkp1 = np.array([a,b,c])

        t[k+1] = tkp1   # speichere Zeitpunkt t_{k+1}
        x[k+1,:] = xkp1 # speichere Loesung x_{k+1}
        tkp1 = tkp1+h   # update t_{k+1} fuer naechsten Schritt
        xkm1 = xk       # update x_{k-1} fuer naechsten Schritt
        xk = xkp1       # update x_{k} fuer naechsten Schritt
#   endfor
    return (x,t)
# endfunction

def main():
    t0 = 0
    tf = 5e-4
    h = 1e-8
    x0 = np.array([0.0,1.0,0.0])
    x1 = np.array([math.sin(h),math.cos(h),-math.sin(h)])
    
    x,t = bdf2(t0,tf,x0,x1,h)

    xma = max(x[:,1])
    xmi = min(x[:,1])

    print('x2_max_min:',xma,xmi)

    # berechne exakte Lösung
    nn = len(t)
    xx = np.ones((nn,len(x0)))
    for i in range(nn):
        xx[i,0] = math.sin(t[i])
        xx[i,1] = math.cos(t[i])
        xx[i,2] = -math.sin(t[i])

    # graphische Darstellung der numerischen Loesung
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(12.5,4))
    ax1.plot(t, x[:,0], color="blue", linewidth=1,linestyle="-", label='$x_1$ num.')
    ax2.plot(t, x[:,1], color="blue", linewidth=1,linestyle="-", label='$x_2$ num.')
    ax3.plot(t, x[:,2], color="blue", linewidth=1,linestyle="-", label='$x_3$ num.')

    # graphischer Vergleich mit der exakten Loesung
    ax1.plot(t, xx[:,0], color="red", linewidth=1,linestyle="--", label='$x_1$ exakt')
    ax2.plot(t, xx[:,1], color="red", linewidth=1,linestyle="--", label='$x_2$ exakt')
    ax3.plot(t, xx[:,2], color="red", linewidth=1,linestyle="--", label='$x_3$ exakt')

    ax1.legend()
    ax2.legend()
    ax3.legend()
    fig.suptitle(f'2-schrittige BDF Lösung für Index 3 DAE $x_1\'=x_2, x_2\'=x_3, x_1=\sin(t)$ mit h={h}')
    plt.savefig("bdf2_index3_smallstepsize.png",dpi=300, bbox_inches='tight')
    plt.show()

if __name__ == '__main__':
    main()

