#!/usr/bin/python
# -*-coding: utf-8-*-

# Darstellung der Loesungen fuer die Gear-Petzold-DAE
# in klassischer Formulierung und in einer Formulierung
# mit einem proper formuliertem Ableitungsterm 

import numpy as np
import scipy.optimize as scopt
import math
import matplotlib.pyplot as plt

# Programm zur Bereitstellung 
# der Funktion f(y,x,t) der
# DAE f(d(x,t)',x,t) = 0.
# 
def f(y,x,t,parameter):
    eta = parameter[0]
    daetype = parameter[1]
    if daetype == 'proper':
        f1 = y + x[1]
        f2 = x[0] + eta*t*x[1] - math.exp(-t)
    else: # daetype = 'classic'
        f1 = y[0] + eta*t*y[1] + (1+eta)*x[1]
        f2 = x[0] + eta*t*x[1] - math.exp(-t)
    return np.array([f1,f2])
# endfunction

# Programm zur Bereitstellung 
# der Funktion d(x,t) der
# DAE f(d(x,t)',x,t) = 0.
# 
def d(x,t,parameter):
    eta = parameter[0]
    daetype = parameter[1]
    if daetype == 'proper':
        d = x[0] + eta*t*x[1]
    else: # daetype = 'classic'
        d = x
    return d
# endfunction

# Exakte Loesung
def exakte_lsg(t,parameter):
    eta = parameter[0]
    x2 = math.exp(-t)
    x1 = (1.0-eta*t)*x2
    return np.array([x1,x2])
# endfunction

# Berechne exakte Loesung auf Intervall
def exakt(t0,tf,x0,h,parameter):
    n = int((tf - t0)/h)
    t = np.linspace(t0,tf,num=n)
    x = np.ones((len(t),len(x0)))

    for i in range(n):
        x[i,:] = exakte_lsg(t[i],parameter)
    # end for
    return (x,t)
# endfunction

# implizites Eulerverfahren
def ieuler(t0,tf,x0,h,parameter):
    n = int((tf - t0)/h)
    m = len(x0)

    t = np.zeros(n+1); x = np.zeros((n+1,m))
    xk = np.zeros(m)
    t[0] = t0
    x[0] = x0
    xk = x0 # starte mit x_{k}:=x0
    tkp1 = t0+h # neuer Zeitpunkt t_{k+1}

    # Die Loesung xkp1=x_{k+1} eines impliziten Euler-Schrittes
    # ist die Nullstelle von ieulerstep_f (nahe x_k).
    def ieulerstep_f(x,xk,tkp1,h,parameter):
        dx = d(x,tkp1,parameter)
        dxk = d(xk,tkp1-h,parameter)
        ykp1 = (dx-dxk)/h
        return f(ykp1,x,tkp1,parameter)
    # end def

    for k in range(n):
        args = (xk,tkp1,h,parameter)

        ## berechne Nullstelle xkp1 von ieulerstep_f
        ## und nehme x_k als Startnaeherung
        xkp1 = scopt.fsolve(ieulerstep_f,xk,args)

        t[k+1] = tkp1   # speichere Zeitpunkt t_{k+1}
        x[k+1,:] = xkp1 # speichere Loesung x_{k+1}
        tkp1 = tkp1+h   # update t_{k+1} fuer naechsten Schritt
        xk = xkp1       # update x_{k} fuer naechsten Schritt
    # end for
    return (x,t)
# end def

# 2-schrittiges BDF Verfahren
def bdf2(t0,tf,x0,x1,h,parameter):
    n = int((tf - t0)/h)
    m = len(x0)

    t = np.zeros(n+1); x = np.zeros((n+1,m))
    xk = np.zeros(m); xkm1 = np.zeros(m)
    t[0] = t0; t[1] = t0+h
    x[0] = x0; x[1] = x1
    xkm1 = x0 # starte mit x_{k-1}:=x0
    xk = x1   # und x_{k}:=x1
    tkp1 = t0+2.0*h # neuer Zeitpunkt t_{k+1}

    # Die Loesung xkp1=x_{k+1} eines BDF-Schrittes
    # ist die Nullstelle von bdf2step_f (nahe x_k).
    def bdf2step_f(x,xk,xkm1,tkp1,h,parameter):
        dx = d(x,tkp1,parameter)
        dxk = d(xk,tkp1-h,parameter)
        dxkm1 = d(xkm1,tkp1-2.0*h,parameter)
        ykp1 = (1.5*dx-2.0*dxk+0.5*dxkm1)/h
        return f(ykp1,x,tkp1,parameter)
    # end def

    for k in range(1,n):
        args = (xk,xkm1,tkp1,h,parameter)

        ## berechne Nullstelle xkp1 von bdf2step_f
        ## und nehme x_k als Startnaeherung
        xkp1 = scopt.fsolve(bdf2step_f,xk,args)

        t[k+1] = tkp1   # speichere Zeitpunkt t_{k+1}
        x[k+1,:] = xkp1 # speichere Loesung x_{k+1}
        tkp1 = tkp1+h   # update t_{k+1} fuer naechsten Schritt
        xkm1 = xk       # update x_{k-1} fuer naechsten Schritt
        xk = xkp1       # update x_{k} fuer naechsten Schritt
    # end for
    return (x,t)
# end def

# 2-stufiges Radau-IIA-Verfahren
def radauIIA(t0,tf,x0,h,parameter):
    n = int((tf - t0)/h)
    m = len(x0)

    t = np.zeros(n+1); x = np.zeros((n+1,m))
    xk = np.zeros(m)
    xk = x0 # starte mit x_k:=x0
    tk = t0 # aktueller Zeitpunkt t_k
    t[0] = t0; x[0] = x0

    beta11 = 1.5
    beta12 = 0.5
    beta21 = -4.5
    beta22 = 2.5
    a1 = 1.0/3.0
    a2 = 1.0

    # Die Stufenloesungen X=(X1,X2) eines Radau-IIA-Schrittes
    # sind eine Nullstelle von radauIIAstep_f (nahe (x_k,x_k)).
    def radauIIAstep_f(X,xk,tk,h,parameter):
        X1 = X[0:m]
        X2 = X[m:2*m]
        Y1 = d(X1,tk+a1*h,parameter)
        Y2 = d(X2,tk+a2*h,parameter)
        yk = d(xk,tk,parameter)
        F1 = f((beta11*(Y1-yk) + beta12*(Y2-yk))/h,X1,tk+a1*h,parameter)
        F2 = f((beta21*(Y1-yk) + beta22*(Y2-yk))/h,X2,tk+a2*h,parameter)

        return np.concatenate((F1,F2))
    # end def 
    for k in range(n):
        args = (xk,tk,h,parameter)

        # berechne Nullstelle von radauIIAstep_f
        # und nehme (x_k,x_k) als Startnaeherung:
        Xstart = np.concatenate((xk,xk))
        X = scopt.fsolve(radauIIAstep_f,Xstart,args)
        xkp1 = X[m:2*m] # Loesung der letzten Stufe s=2

        tk = tk+h       # update t_{k} fuer naechsten Schritt
        xk = xkp1       # update x_{k} fuer naechsten Schritt
        t[k+1] = tk     # speichere aktuellen Zeitpunkt t_{k+1}
        x[k+1,:] = xk   # speichere aktuelle Loesung x_{k+1}
    # endfor
    return (x,t)
# end def 

def main():
    t0 = 0.0
    tf = 4.0
    h = 1e-1

    # setze Parameter fuer die graphische Darstellung
    plt.rcParams.update({'font.size': 12, 'axes.labelsize': 16})
    plt.rcParams.update({'xtick.labelsize': 12, 'ytick.labelsize': 12})
    plt.rcParams.update({'figure.subplot.top': 0.92,'figure.subplot.bottom': 0.08})
    plt.rcParams.update({'figure.subplot.hspace': 0.3,'figure.subplot.hspace': 0.3})

    # berechne die numerischen Loesungen + graphische Darstellung
    for daetype in ['classic','proper']:
        # erzeuge Bild mit 4 Teilbildern
        fig, axs = plt.subplots(2, 2, figsize=(12.5,8))
        #plt.rcParams.update('xtick', labelsize=16)
        i_ind = [0,0,1,1]
        j_ind = [0,1,0,1]
        k = 0

        for eta in [-0.20,-0.3,-0.4,-0.6]:
            parameter = (eta,daetype)
            x0 = exakte_lsg(t0,parameter)
            x1 = exakte_lsg(t0+h,parameter)
            x_exakt, t_exakt = exakt(t0,tf,x0,h,parameter)
            x_ieuler, t_ieuler = ieuler(t0,tf,x0,h,parameter)
            x_bdf2, t_bdf2 = bdf2(t0,tf,x0,x1,h,parameter)
            x_radauIIA, t_radauIIA = radauIIA(t0,tf,x0,h,parameter)
            i = i_ind[k]; j = j_ind[k]; k = k+1
            axs[i,j].plot(t_radauIIA, x_radauIIA[:,0], color="green", linewidth=2,linestyle="-", label='RadauIIA')
            axs[i,j].plot(t_ieuler, x_ieuler[:,0], color="black", linewidth=2,linestyle="-.", label='impl. Euler')
            axs[i,j].plot(t_bdf2, x_bdf2[:,0], color="red", linewidth=2,linestyle="--", label='BDF-2')
            axs[i,j].plot(t_exakt, x_exakt[:,0], color="blue", linewidth=2,linestyle="dotted", label='exakt')
            axs[i,j].set_ylim((0.0,1.0))
            axs[i,j].set_xlabel('t')
            axs[i,j].set_ylabel('$x_1$')
            axs[i,j].set_title(f'$\eta$ = {eta}, h = {h}')
            axs[i,j].legend()
    
        #fig.suptitle(f'DAE $x_1\'+(\eta t x_2)\'=0, x_1+\eta t x_2 = \exp(-t)$ mit h = {h}')
        plt.savefig("Gear-Petzold-DAE_"+str(daetype)+".png",dpi=300, bbox_inches='tight')
        plt.show()
# end def

if __name__ == '__main__':
    main()
